import os

import numpy as np
import pandas as pd

current_directory = os.path.abspath(os.path.dirname(__file__))


class Namespace:
    pass


class ResultsTimeSeries:
    """
    Results are from statsmodels.
    """

    @staticmethod
    def correlation():
        obj = Namespace()
        obj.acf = np.loadtxt(fname=os.path.join(current_directory, 'acf.csv'), dtype=float, delimiter=",")
        obj.pacf = np.loadtxt(fname=os.path.join(current_directory, 'pacf.csv'), dtype=float, delimiter=",")
        return obj

    @staticmethod
    def signs_statistics(column_name: str) -> Namespace:
        obj = Namespace()
        obj.percentiles = (50.0, 75.0, 95.0, 99.0, 99.9)
        index = ["size", "pct_buy (%)", "mean_nb_consecutive_values", "std_nb_consecutive_values",
                 "Q50.0_nb_consecutive_values", "Q75.0_nb_consecutive_values", "Q95.0_nb_consecutive_values",
                 "Q99.0_nb_consecutive_values", "Q99.9_nb_consecutive_values"]
        statistics = [1000.0, 64.0, 3.7037037037037037, 7.845332723462921, 2.0, 3.0, 13.0, 35.13000000000005, 78.00600000000134]
        obj.stats_df = pd.DataFrame(data=statistics, columns=[column_name], index=index)

        return obj



from enum import Enum

import numpy as np
import pytest

from tradeflow.exceptions import EnumValueException
from tradeflow.general_utils import check_condition, get_enum_values, \
    check_enum_value_is_valid, \
    is_value_within_interval_exclusive


class MyEnumWithoutNone(Enum):
    ENUM_1 = "enum_value_1"
    ENUM_2 = "enum_value_2"


class MyEnumWithNone(Enum):
    ENUM_1 = "enum_value_1"
    ENUM_2 = "enum_value_2"
    NONE = None


@pytest.fixture
def enum_without_none():
    yield MyEnumWithoutNone


@pytest.fixture
def enum_with_none():
    yield MyEnumWithNone


def test_check_condition():
    assert check_condition(condition=True, exception=Exception("Exception message")) is None


def test_check_condition_should_raise_exception():
    with pytest.raises(Exception) as ex:
        check_condition(condition=False, exception=Exception("Exception message"))
    assert str(ex.value) == "Exception message"


@pytest.mark.parametrize("enum,expected_enum_values", [
    ("enum_without_none", ["enum_value_1", "enum_value_2"]),
    ("enum_with_none", ["enum_value_1", "enum_value_2", None])
])
def test_get_enum_values(enum, expected_enum_values, request):
    actual_enum_values = get_enum_values(enum_obj=request.getfixturevalue(enum))
    assert np.array_equal(expected_enum_values, actual_enum_values)


@pytest.mark.parametrize("enum,enum_value,is_none_valid,expected_enum", [
    ("enum_without_none", "enum_value_1", True, MyEnumWithoutNone.ENUM_1),
    ("enum_without_none", "enum_value_1", False, MyEnumWithoutNone.ENUM_1),
    ("enum_without_none", None, True, None),
    ("enum_with_none", None, True, None),
])
def test_check_enum_value_is_valid(enum, enum_value, is_none_valid, expected_enum, request):
    enum = request.getfixturevalue(enum)
    assert check_enum_value_is_valid(enum_obj=enum, value=enum_value, is_none_valid=is_none_valid, parameter_name="enum_value") == expected_enum


@pytest.mark.parametrize("enum,enum_value,is_none_valid,expected_valid_values", [
    ("enum_without_none", "enum_3", True, "['enum_value_1', 'enum_value_2']"),
    ("enum_without_none", "enum_3", False, "['enum_value_1', 'enum_value_2']"),
    ("enum_without_none", None, False, "['enum_value_1', 'enum_value_2']"),
    ("enum_with_none", None, False, "['enum_value_1', 'enum_value_2', None]")
])
def test_check_enum_value_is_valid_should_raise_exception(enum, enum_value, is_none_valid, expected_valid_values, request):
    expected_exception_message = f"The value '{enum_value}' for enum_value is not valid, it must be among {expected_valid_values} or None if it is valid."
    enum = request.getfixturevalue(enum)
    with pytest.raises(EnumValueException) as ex:
        check_enum_value_is_valid(enum_obj=enum, value=enum_value, is_none_valid=is_none_valid, parameter_name="enum_value")

    assert str(ex.value) == expected_exception_message


@pytest.mark.parametrize("value,lower_bound,upper_bound,expected", [
    (1, 0, 2, True),
    (1, 0.9, 1.5, True),
    (1.5, 0.9, 1.6, True),
    (1, 1, 2, False),
    (1, 1.1, 1.5, False),
    (1.0, 1.0, 1.5, False),
])
def test_is_value_within_interval_exclusive(value, lower_bound, upper_bound, expected):
    assert is_value_within_interval_exclusive(value=value, lower_bound=lower_bound, upper_bound=upper_bound) == expected



import os

import numpy as np

from tradeflow.constants import FitMethodAR

current_directory = os.path.abspath(os.path.dirname(__file__))


class Namespace:
    pass


class ResultsAR:
    """
    Results are from statsmodels.
    """

    @staticmethod
    def parameters_order_6(method: str) -> Namespace:
        obj = Namespace()
        match method:
            case FitMethodAR.YULE_WALKER.value:
                obj.constant_parameter = 0
                obj.parameters = [0.20793670441358317, 0.15625334330632215, 0.08328570101676176, 0.10762268507210443, 0.12228963258158163, 0.07896963026086244]
            case FitMethodAR.OLS_WITH_CST.value:
                obj.constant_parameter = 0.06400645100573674
                obj.parameters = [0.2091518029574414, 0.15836060267986327, 0.0834257462768201, 0.10653755478818688, 0.12165411140031872, 0.07969039271249247]

        return obj

    @staticmethod
    def simulated_signs(fit_method: str) -> Namespace:
        obj = Namespace()
        match fit_method:
            case FitMethodAR.YULE_WALKER.value:
                obj.simulation = np.loadtxt(fname=os.path.join(current_directory, 'simulated_signs_yule_walker.csv'), dtype=float, delimiter=",")
            case FitMethodAR.OLS_WITH_CST.value:
                obj.simulation = np.loadtxt(fname=os.path.join(current_directory, 'simulated_signs_ols_with_cst.csv'), dtype=float, delimiter=",")

        return obj



import logging
import sys

from tradeflow.constants import Logger

logging.basicConfig(format=Logger.FORMAT, level=Logger.LEVEL)


def get_logger(name: str, print_to_stdout=True) -> logging.Logger:
    logger = logging.getLogger(name)
    if print_to_stdout:
        handler = create_stdout_handler()
        logger.addHandler(handler)
    return logger


def create_stdout_handler() -> logging.StreamHandler:
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(fmt=Logger.FORMAT))
    handler.setLevel(level=Logger.LEVEL)
    return handler



from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
from matplotlib.axes import Axes
from numpy.testing import assert_equal, assert_almost_equal
from pandas.testing import assert_frame_equal

from tradeflow.datasets import signs
from tradeflow.exceptions import IllegalNbLagsException, IllegalValueException
from tradeflow.tests.results.results_time_series import ResultsTimeSeries
from tradeflow.time_series import TimeSeries

signs_data = signs.load()


@pytest.fixture
def time_series_signs():
    TimeSeries.__abstractmethods__ = set()
    time_series = TimeSeries(signs=signs_data)
    time_series._order = 6
    return time_series


@pytest.fixture
def time_series_signs_non_stationary():
    TimeSeries.__abstractmethods__ = set()
    time_series = TimeSeries(signs=[-1] * 500 + [1] * 500)
    time_series._order = 1
    return time_series


class TestCalculateAcf:

    @pytest.mark.parametrize("nb_lags", [-1, 0, 1000, 1500])
    def test_calculate_acf_should_raise_exception(self, time_series_signs, nb_lags):
        expected_exception_message = f"Can only calculate the autocorrelation function with a number of lags positive and lower than the time series length (requested number of lags {nb_lags} should be < 1000)."
        with pytest.raises(IllegalNbLagsException) as ex:
            time_series_signs.calculate_acf(nb_lags=nb_lags)

        assert str(ex.value) == expected_exception_message

    @pytest.mark.parametrize("nb_lags", [1, 200, 999])
    def test_calculate_acf(self, time_series_signs, nb_lags):
        actual_acf = time_series_signs.calculate_acf(nb_lags=nb_lags)
        assert len(actual_acf) == nb_lags + 1

        expected_acf = ResultsTimeSeries.correlation().acf[:nb_lags + 1]
        assert_almost_equal(actual=actual_acf, desired=expected_acf, decimal=10)


class TestCalculatePacf:

    @pytest.mark.parametrize("nb_lags", [-1, 0, 500, 750])
    def test_calculate_pacf_should_raise_exception_when_invalid_nb_lags(self, time_series_signs, nb_lags):
        expected_exception_message = f"Can only calculate the partial autocorrelation function with a number of lags positive and lower than 50% of the time series length (requested number of lags {nb_lags} should be < 500)."
        with pytest.raises(IllegalNbLagsException) as ex:
            time_series_signs.calculate_pacf(nb_lags=nb_lags, alpha=0.05)

        assert str(ex.value) == expected_exception_message

    @pytest.mark.parametrize("alpha", [-0.05, 1.05])
    def test_calculate_pacf_should_raise_exception_when_invalid_alpha(self, time_series_signs, alpha):
        expected_exception_message = f"Alpha {alpha} is invalid, it must be in the interval [0, 1]"
        with pytest.raises(IllegalValueException) as ex:
            time_series_signs.calculate_pacf(nb_lags=25, alpha=alpha)

        assert str(ex.value) == expected_exception_message

    @pytest.mark.parametrize("nb_lags,alpha", [(1, 0.05), (200, None), (499, 0.05)])
    def test_calculate_pacf(self, time_series_signs, nb_lags, alpha):
        actual_pacf = time_series_signs.calculate_pacf(nb_lags=nb_lags, alpha=alpha)
        actual_pacf = actual_pacf[0] if alpha is not None else actual_pacf
        assert len(actual_pacf) == nb_lags + 1

        expected_pacf = ResultsTimeSeries.correlation().pacf[:nb_lags + 1]
        assert_almost_equal(actual=actual_pacf, desired=expected_pacf, decimal=10)


class TestTimeSeriesStationarity:

    @pytest.mark.parametrize("regression", ["c", "ct", "ctt", "n"])
    def test_time_series_should_be_stationary(self, time_series_signs, regression):
        assert time_series_signs._is_time_series_stationary(significance_level=0.05, regression=regression)

    @pytest.mark.parametrize("regression", ["c", "ct", "ctt", "n"])
    def test_time_series_should_be_non_stationary(self, time_series_signs_non_stationary, regression):
        assert not time_series_signs_non_stationary._is_time_series_stationary(significance_level=0.05, regression=regression)


class TestSimulationSummary:

    def test_simulation_summary(self, time_series_signs):
        time_series_signs._simulation = time_series_signs._signs
        time_series_signs._order = 6

        actual_simulation_summary_df = time_series_signs.simulation_summary(plot=False, log_scale=False, percentiles=(50.0, 75.0, 95.0, 99.0, 99.9))

        expected_training_stats_df = ResultsTimeSeries.signs_statistics(column_name="Training").stats_df
        expected_simulation_stats_df = ResultsTimeSeries.signs_statistics(column_name="Simulation").stats_df
        expected_simulation_summary_df = pd.concat([expected_training_stats_df, expected_simulation_stats_df],
                                                   axis=1).round(decimals=2)

        assert_frame_equal(left=actual_simulation_summary_df, right=expected_simulation_summary_df, check_dtype=True,
                           check_index_type=True, check_names=True, check_exact=True, obj="stats")

    def test_compute_signs_statistics(self):
        results_signs_stats = ResultsTimeSeries.signs_statistics(column_name="Test signs")
        actual_stats_df = TimeSeries._compute_signs_statistics(signs=signs_data, column_name="Test signs",
                                                               percentiles=results_signs_stats.percentiles)

        expected_stats_df = results_signs_stats.stats_df
        assert_frame_equal(left=actual_stats_df, right=expected_stats_df, check_dtype=True, check_index_type=True,
                           check_names=True,
                           check_exact=True, obj="stats")

    @pytest.mark.parametrize("signs,expected_series", [
        ([1., 1., -1., -1., -1., 1., 1.], [2, 3, 2]),
        ([-1., 1., -1., 1., 1.], [1, 1, 1, 2]),
        ([1., -1., 1., -1.], [1, 1, 1, 1]),
        ([-1., 1., -1., -1., -1., -1., 1., 1., 1., 1., 1., -1.], [1, 1, 4, 5, 1]),
        ([1., 2., 2., 3., 3., 3., 4., 4., 4., 4.], [1, 2, 3, 4])
    ])
    def test_compute_series_nb_consecutive_values(self, signs, expected_series):
        actual_series = TimeSeries._compute_series_nb_consecutive_signs(signs=signs)
        assert_equal(actual=actual_series, desired=expected_series)

    @pytest.mark.parametrize("signs,expected_buy_pct", [
        ([1., -1., 1., 1., -1.], 100 * 3 / 5),
        ([1., 1., 1., 1., 1.], 100.0),
        ([-1., -1., -1., -1., -1.], 0.0),
        ([-1., -1., -1., -1., 1.], 100 * 1 / 5)
    ])
    def test_percentage_buy(self, signs, expected_buy_pct):
        assert TimeSeries._percentage_buy(signs=signs) == expected_buy_pct


class TestPlot:

    @staticmethod
    def check_axe_values_training_vs_simulation(axe: Axes, training_values: np.ndarray, simulation_values: np.ndarray,
                                                order: int, title: str, y_scale: str, x_lim: Tuple[float, float],
                                                y_lim: Tuple[float, float] | None = None):
        assert_almost_equal(actual=axe.lines[0].get_xydata()[:, 1], desired=training_values, decimal=10)
        assert axe.lines[0].get_label() == "Training"

        assert_almost_equal(actual=axe.lines[1].get_xydata()[:, 1], desired=simulation_values, decimal=10)
        assert axe.lines[1].get_label() == "Simulation"

        assert np.all([x == order for x in axe.lines[2].get_xydata()[:, 0]])
        assert axe.get_title() == title
        assert axe.get_xlabel() == "Lag"

        assert axe.get_xscale() == "linear"
        assert axe.get_yscale() == y_scale

        assert axe.get_xlim() == x_lim
        if y_lim is not None:
            assert axe.get_ylim() == y_lim

    @pytest.mark.parametrize("log_scale", [True, False])
    def test_build_fig_corr_training_vs_simulation(self, time_series_signs, log_scale):
        order = time_series_signs._order
        time_series_signs._simulation = time_series_signs._signs
        y_scale = "log" if log_scale else "linear"

        fig = time_series_signs._build_fig_corr_training_vs_simulation(log_scale=log_scale)

        acf_axe = fig.get_axes()[0]
        expected_acf = ResultsTimeSeries.correlation().acf[:2 * order + 1]
        expected_acf_title = f"ACF plot for training and simulated time series ({y_scale} scale)"
        self.check_axe_values_training_vs_simulation(axe=acf_axe, training_values=expected_acf,
                                                     simulation_values=expected_acf,
                                                     order=order, title=expected_acf_title, y_scale=y_scale,
                                                     x_lim=(-1.0, 2 * order))

        pacf_axe = fig.get_axes()[1]
        expected_pacf = ResultsTimeSeries.correlation().pacf[:2 * order + 1]
        expected_pacf_title = f"PACF plot for training and simulated time series ({y_scale} scale)"
        self.check_axe_values_training_vs_simulation(axe=pacf_axe, training_values=expected_pacf,
                                                     simulation_values=expected_pacf, order=order,
                                                     title=expected_pacf_title,
                                                     y_scale=y_scale, x_lim=(-1.0, 2 * order))

    @pytest.mark.parametrize("log_scale", [True, False])
    def test_fill_axe_training_vs_simulation(self, log_scale):
        training_values = np.array([1, 2, 3, 2, 2.5])
        simulation_values = np.array([1.3, 2.1, 2.9, 2.2, 2.4])
        order = 2

        TimeSeries.__abstractmethods__ = set()
        time_series = TimeSeries(signs=training_values)
        time_series._simulation = simulation_values
        time_series._order = order

        title = "Test plot training vs simulation"

        fig, axe = plt.subplots(1, 1, figsize=(8, 4))
        time_series._fill_axe_training_vs_simulation(axe=axe, training=training_values,
                                                     simulation=simulation_values,
                                                     order=order, title=title, log_scale=log_scale)

        y_scale = "log" if log_scale else "linear"
        expected_title = f"{title} ({y_scale} scale)"
        self.check_axe_values_training_vs_simulation(axe=axe, training_values=training_values,
                                                     simulation_values=simulation_values, order=order,
                                                     title=expected_title,
                                                     y_scale=y_scale,
                                                     x_lim=(-1.0, 4.0), y_lim=(1.0, 3.1))



import pytest
from numpy.testing import assert_equal, assert_almost_equal

from tradeflow.ar_model import AR
from tradeflow.datasets import signs
from tradeflow.exceptions import IllegalNbLagsException, EnumValueException, \
    IllegalValueException, ModelNotFittedException, NonStationaryTimeSeriesException
from tradeflow.tests.results.results_ar_model import ResultsAR

signs_data = signs.load()


@pytest.fixture
def ar_model_with_max_order_6():
    ar_model = AR(signs=signs_data, max_order=6, order_selection_method=None, information_criterion=None)
    return ar_model


@pytest.fixture
def ar_model_non_stationary_with_max_order_1():
    ar_model = AR(signs=[-1] * 500 + [1] * 500, max_order=1, order_selection_method=None, information_criterion=None)
    return ar_model


class TestInit:

    @pytest.mark.parametrize("max_order", [500, 1000])
    def test_init_max_order_should_raise_exception_when_invalid_max_order(self, max_order):
        with pytest.raises(IllegalNbLagsException) as ex:
            AR(signs=signs_data, max_order=max_order, order_selection_method=None, information_criterion=None)

        assert str(ex.value) == f"{max_order} is not valid for 'max_order', it must be positive and lower than 50% of the time series length (< 500)."

    def test_init_should_raise_exception_when_invalid_order_selection_method(self):
        with pytest.raises(EnumValueException) as ex:
            AR(signs=signs_data, max_order=6, order_selection_method="invalid_order_selection_method", information_criterion="aic")

        assert str(ex.value) == "The value 'invalid_order_selection_method' for order_selection_method is not valid, it must be among ['information_criterion', 'pacf'] or None if it is valid."

    @pytest.mark.parametrize("order_selection_method,information_criterion", [
        ("information_criterion", "invalid_ic"),
        ("information_criterion", None)
    ])
    def test_init_should_raise_exception_when_invalid_information_criterion(self, order_selection_method, information_criterion):
        expected_exception_message = f"The value '{information_criterion}' for information_criterion is not valid, it must be among ['aic', 'bic', 'hqic'] or None if it is valid."
        with pytest.raises(EnumValueException) as ex:
            AR(signs=signs_data, max_order=6, order_selection_method=order_selection_method, information_criterion=information_criterion)

        assert str(ex.value) == expected_exception_message


class TestInitMaxOrder:

    @pytest.mark.parametrize("max_order,expected_max_order", [
        (25, 25),
        (None, 22)  # Schwert (1989)
    ])
    def test_init_max_order(self, ar_model_with_max_order_6, max_order, expected_max_order):
        assert ar_model_with_max_order_6._init_max_order(max_order=max_order) == expected_max_order

    @pytest.mark.parametrize("max_order", [0, 500, 1000])
    def test_init_max_order_should_raise_exception_when_invalid_max_order(self, ar_model_with_max_order_6, max_order):
        expected_exception_message = f"{max_order} is not valid for 'max_order', it must be positive and lower than 50% of the time series length (< 500)."
        with pytest.raises(IllegalNbLagsException) as ex:
            ar_model_with_max_order_6._init_max_order(max_order=max_order)

        assert str(ex.value) == expected_exception_message


class TestFit:

    @pytest.mark.parametrize("method", ["yule_walker", "ols_with_cst"])
    def test_fit(self, ar_model_with_max_order_6, method):
        ar_model_with_max_order_6.fit(method=method)

        expected_parameters_results = ResultsAR.parameters_order_6(method=method)
        assert_almost_equal(actual=ar_model_with_max_order_6._constant_parameter, desired=expected_parameters_results.constant_parameter, decimal=10)
        assert_almost_equal(actual=ar_model_with_max_order_6._parameters, desired=expected_parameters_results.parameters, decimal=10)

    @pytest.mark.parametrize("method", ["invalid_method", None])
    def test_fit_should_raise_exception_when_invalid_method(self, ar_model_with_max_order_6, method):
        expected_exception_message = f"The value '{method}' for method is not valid, it must be among ['yule_walker', 'ols_with_cst'] or None if it is valid."
        with pytest.raises(EnumValueException) as ex:
            ar_model_with_max_order_6.fit(method=method)

        assert str(ex.value) == expected_exception_message

    @pytest.mark.parametrize("method", ["yule_walker", "ols_with_cst"])
    def test_fit_should_raise_exception_when_time_series_non_stationary(self, ar_model_non_stationary_with_max_order_1, method):
        with pytest.raises(NonStationaryTimeSeriesException) as ex:
            ar_model_non_stationary_with_max_order_1.fit(method="yule_walker")

        assert str(ex.value) == "The time series must be stationary to be fitted."


class TestSelectOrder:

    @pytest.mark.parametrize("max_order,order_selection_method,information_criterion,expected_order", [
        (25, "information_criterion", "aic", 6), (4, "information_criterion", "aic", 4),
        (50, "information_criterion", "bic", 5), (3, "information_criterion", "bic", 3),
        (25, "information_criterion", "hqic", 6), (2, "information_criterion", "hqic", 2),
        (499, "pacf", "hqic", 6), (1, "pacf", "hqic", 1)
    ])
    def test_select_order_with_selection_method(self, max_order, order_selection_method, information_criterion, expected_order):
        ar_model = AR(signs=signs_data, max_order=max_order, order_selection_method=order_selection_method, information_criterion=information_criterion)
        assert ar_model._max_order == max_order

        ar_model._select_order()
        assert ar_model._order == expected_order

    @pytest.mark.parametrize("max_order,information_criterion,expected_order", [
        (25, None, 25),
        (25, "aic", 25),
        (None, None, 22),  # Schwert (1989)
        (None, "aic", 22)  # Schwert (1989)
    ])
    def test_select_order_without_selection_method(self, max_order, information_criterion, expected_order):
        ar_model = AR(signs=signs_data, max_order=max_order, order_selection_method=None, information_criterion=information_criterion)
        ar_model._select_order()
        assert ar_model._order == expected_order == ar_model._max_order


class TestSimulate:

    @pytest.mark.parametrize("method", ["yule_walker", "ols_with_cst"])
    @pytest.mark.parametrize("size", [50, 1000])
    def test_simulate(self, ar_model_with_max_order_6, method, size):
        actual_simulation = ar_model_with_max_order_6.fit(method=method).simulate(size=size, seed=1)

        expected_signs = ResultsAR.simulated_signs(fit_method=method)
        assert len(actual_simulation) == size
        assert_equal(actual=actual_simulation, desired=expected_signs.simulation[:size])

    @pytest.mark.parametrize("size", [-50, 0])
    def test_simulate_should_raise_exception_when_invalid_size(self, ar_model_with_max_order_6, size):
        with pytest.raises(IllegalValueException) as ex:
            ar_model_with_max_order_6.fit("yule_walker").simulate(size=size)

        assert str(ex.value) == f"The size '{size}' for the time series to be simulated is not valid, it must be greater than 0."

    def test_simulate_should_raise_exception_when_model_not_fitted(self, ar_model_with_max_order_6):
        with pytest.raises(ModelNotFittedException) as ex:
            ar_model_with_max_order_6.simulate(size=50)

        assert str(ex.value) == "The model has not yet been fitted. Fit the model first by calling 'fit()'."



from statsmodels.datasets import utils as du


def load():
    """
    Load trade signs data and returns a numpy ndarray.

    Returns
    -------
    Dataset
    """
    return load_pandas()


def load_pandas():
    """
    Load trade signs data and returns a numpy.ndarray.

    Returns
    -------
    Dataset
    """
    return _get_data().values


def _get_data():
    data = du.load_csv(__file__, 'signs.csv')
    data = data.iloc[:, 1]
    return data.astype(float)
